import numpy as np
import traitlets
import vaex.ml.state
import logging
import vaex
import vaex.serialize
from numba import jit
from . import generate
# vaex.set_log_level_debug()

logger_km = logging.getLogger('vaex.ml.kmeans')


def Matrix(type=traitlets.CFloat):
    return traitlets.List(traitlets.List(type()))


# @jit('void(double[:], double[:], double[:])', nopython=True, nogil=True)
@jit(nopython=True, nogil=True, cache=True)
def distances_square(result, centroids, *blocks):
    """
    Function under test.
    """
    k = centroids.shape[0]
    dimensions = centroids.shape[1]
    N = result.shape[1]
    for i in range(k):
        for d in range(dimensions):
            for j in range(N):
                result[i, j] += (blocks[d][j] - centroids[i][d])**2


@jit(nopython=True, nogil=True, cache=True)
def centroid_stats(centroids, counts, sumpos, inertia, *blocks):
    # classes = np.argmin(distances_sq, axis=0)
    # distances_sq = np.choose(classes, distances_sq)
    # inertia = (distances_sq).sum()
    # inertia = 0
    N = blocks[0].shape[0]
    runs = sumpos.shape[0]
    clusters = sumpos.shape[1]
    dimensions = sumpos.shape[2]
    for j in range(N):
        for run in range(runs):
            # if done[run]: ## TODO: bug in numba? seems to crash if included
                best_distance = 1e100
                best_class = 10000
                for cluster in range(clusters):
                    distance = 0
                    for d in range(dimensions):
                        distance += (centroids[run, cluster, d] - blocks[d][j])**2
                    if distance < best_distance:
                        best_class = cluster
                        best_distance = distance
                cls = best_class  # classes[j]
                inertia[run] += best_distance
                for d in range(dimensions):
                    sumpos[run, cls, d] += blocks[d][j]
                counts[run, cls] += 1
#    return inertia
        # for i in range(k):


@vaex.serialize.register
@generate.register
class KMeans(vaex.ml.state.HasState):
    '''The KMeans clustering algorithm.

    Example:

    >>> import vaex.ml
    >>> import vaex.ml.cluster
    >>> df = vaex.ml.datasets.load_iris()
    >>> features = ['sepal_width', 'petal_length', 'sepal_length', 'petal_width']
    >>> cls = vaex.ml.cluster.KMeans(n_clusters=3, features=features, init='random', max_iter=10)
    >>> cls.fit(df)
    >>> df = cls.transform(df)
    >>> df.head(5)
     #    sepal_width    petal_length    sepal_length    petal_width    class_    prediction_kmeans
     0            3               4.2             5.9            1.5         1                    2
     1            3               4.6             6.1            1.4         1                    2
     2            2.9             4.6             6.6            1.3         1                    2
     3            3.3             5.7             6.7            2.1         2                    0
     4            4.2             1.4             5.5            0.2         0                    1
    '''
    features = traitlets.List(traitlets.Unicode(), help='List of features to cluster.')
    n_clusters = traitlets.CInt(default_value=2, help='Number of clusters to form.')
    init = traitlets.Union([Matrix(), traitlets.Unicode()], default_value='random', help='Method for initializing the centroids.')
    n_init = traitlets.CInt(default_value=1, help='Number of centroid initializations. \
                                                   The KMeans algorithm will be run for each initialization, \
                                                   and the final results will be the best output of the n_init \
                                                   consecutive runs in terms of inertia.')
    max_iter = traitlets.CInt(default_value=300, help='Maximum number of iterations of the KMeans algorithm for a single run.')
    random_state = traitlets.CInt(default_value=None, allow_none=True, help='Random number generation for centroid initialization. \
                                                                             If an int is specified, the randomness becomes deterministic.')
    verbose = traitlets.CBool(default_value=False, help='If True, enable verbosity mode.')
    cluster_centers = traitlets.List(traitlets.List(traitlets.CFloat()), help='Coordinates of cluster centers.')
    inertia = traitlets.CFloat(default_value=None, allow_none=True, help='Sum of squared distances of samples to their closest cluster center.')
    prediction_label = traitlets.Unicode(default_value='prediction_kmeans', help='The name of the virtual column that houses the cluster labels for each point.')

    def __call__(self, *blocks):
        return self._calculate_classes(*blocks)

    def _calculate_distances_squared(self, *blocks):
        N = len(blocks[0])  # they are all the same length
        centroids = np.array(self.cluster_centers)
        k = centroids.shape[0]
        dimensions = centroids.shape[1]
        distances_sq = np.zeros((k, N))
        if 1:
            distances_square(distances_sq, centroids, *blocks)
        else:
            for d in range(dimensions):
                for i in range(k):
                    distances_sq[i] += (blocks[d] - centroids[i][d])**2
        return distances_sq

    def _calculate_classes(self, *blocks):
        distances_sq = self._calculate_distances_squared(*blocks)
        classes = np.argmin(distances_sq, axis=0)
        return classes

    def generate_cluster_centers_random(self, dataframe, rng):
        indices = rng.randint(0, len(dataframe), self.n_clusters)
        return [[dataframe.evaluate(feature, i1=i, i2=i+1)[0] for feature in self.features] for i in indices]

    def transform(self, dataframe):
        '''
        Label a DataFrame with a fitted KMeans model.

        :param dataframe: A vaex DataFrame.

        :returns copy: A shallow copy of the DataFrame that includes the cluster labels.
        :rtype: DataFrame
        '''
        copy = dataframe.copy()
        lazy_function = copy.add_function('kmean_predict_function', self, unique=True)
        expression = lazy_function(*self.features)
        copy.add_virtual_column(self.prediction_label, expression, unique=False)
        return copy

    def fit(self, dataframe):
        '''
        Fit the KMeans model to the dataframe.

        :param dataframe: A vaex DataFrame.
        '''
        if self.init == 'random':
            rng = np.random.RandomState(self.random_state)
            self.run_cluster_centers = [self.generate_cluster_centers_random(dataframe, rng) for k in range(self.n_init)]
        else:
            if self.n_init > 1:
                print("WARNING: n_init > 1 , but init given, only doing one run")
            self.run_cluster_centers = [self.init]
        done = [False] * len(self.run_cluster_centers)
        alldone = False
        first = True
        previous_inertias = None
        iteration = 0
        inertias_list = []
        while not alldone:
            new_centers, inertias = self._find_centers_and_inertias(dataframe, done)
            if self.verbose:
                inertia_msges = [('--' if d else '{: 3}'.format(k)) for k, d in zip(inertias, done)]
                if len(inertias) == 1:
                    inertia_msg = inertia_msges[0]
                else:
                    inertia_msg = " | ".join(inertia_msges)
                print('Iteration {: 4}, inertia {}'.format(iteration, inertia_msg))
            if not first:  # we can only to a check after the second iteration
                done = self._is_done(previous_inertias, inertias)
                alldone = np.all(done)
            else:
                first = False
            iteration += 1
            if (iteration >= self.max_iter):
                logger_km.debug('reached max iterations: %s', self.max_iter)
                alldone = True
            previous_inertias = inertias
            inertias_list.append(inertias)
            self.run_cluster_centers = new_centers
        best_run = np.argmin(inertias)
        self.inertia = inertias[best_run]
        self.inertias = np.array(inertias_list)[:, best_run]
        self.cluster_centers = new_centers[best_run]

    def _is_done(self, inertias1, inertias2):
        diffs = [(inertia1 - inertia2) for inertia1, inertia2 in zip(inertias1, inertias2)]
        return [(diff < 1e-3) for diff in diffs]

    def _find_centers_and_inertias(self, dataframe, done):
        done = np.array(done, dtype=np.int8)
        centroids = np.array(self.run_cluster_centers)
        runs = centroids.shape[0]
        clusters = centroids.shape[1]
        dimensions = centroids.shape[2]
        # print("k =", k)
        assert dimensions == len(self.features), "nr of dimensions for centroid should equal nr of features"

        def map(*blocks):  # this will be called with a chunk of the data
            sumpos = np.zeros((runs, clusters, dimensions))
            counts = np.zeros((runs, clusters))
            inertia = np.zeros((runs))
            if 1:
                centroid_stats(centroids, counts, sumpos, inertia,  *blocks)
            else:
                # this is the pure python code
                # although not made for multiple runs yet
                distances_sq = self._calculate_distances_squared(*blocks)
                classes = np.argmin(distances_sq, axis=0)
                distances_sq = np.choose(classes, distances_sq)
                inertia = (distances_sq).sum()
                for i in range(k):
                    mask = classes == i
                    counts[i] = np.sum(mask)
                    for d in range(dimensions):
                        sumpos[i, d] += blocks[d][mask].sum()
            return sumpos, counts, inertia

        def reduce(x, y):
            sumpos1, counts1, inertia1 = x
            sumpos2, counts2, inertia2 = y
            return sumpos1 + sumpos2, counts1 + counts2, inertia1 + inertia2
        sumpos, counts, inertia = dataframe.map_reduce(map, reduce, self.features)
        means = (sumpos/counts[:, :, np.newaxis])
        return means.tolist(), inertia



import base64
import tempfile

import traitlets
import numpy as np
import annoy

import vaex
import vaex.serialize
from vaex.ml import state
from .. import generate


@vaex.serialize.register
@generate.register
class ANNOYModel(state.HasState):
    features = traitlets.List(traitlets.Unicode(), help='List of features to use.')
    n_trees = traitlets.Int(default_value=10, help="Number of trees to build.")
    search_k = traitlets.Int(default_value=-1, help='Jovan?')
    n_neighbours = traitlets.Int(default_value=10, help='Now many neighbours')
    metric = traitlets.Enum(values=['euclidean', 'manhattan', 'angular', 'hamming', 'dot'], default_value='euclidean', help='Metric to use for distance calculations')
    predcition_name = prediction_name = traitlets.Unicode(default_value='annoy_prediction', help='Output column name for the neighbours when transforming a DataFrame')

    def __call__(self, *args):
        data2d = np.vstack([arg.astype(np.float64) for arg in args]).T.copy()
        result = []
        for i in range(len(data2d)):
            result.append(self.index_builder.get_nns_by_vector(n=self.n_neighbours, vector=data2d[i], search_k=self.search_k))
        return np.array(result)

    def fit(self, dataframe):
        self.index_builder = annoy.AnnoyIndex(len(self.features), metric=self.metric)
        for i in range(len(dataframe)):
            self.index_builder.add_item(i, dataframe[self.features][i])
        self.index_builder.build(self.n_trees)

    def transform(self, dataframe):
        copy = dataframe.copy()
        lazy_function = copy.add_function('annoy_get_nns_by_vector_function', self, unique=True)
        expression = lazy_function(*self.features)
        copy.add_virtual_column(self.prediction_name, expression, unique=False)
        return copy

    def predict(self, dataframe, n_neighbours=None, search_k=None):
        search_k = search_k or self.search_k
        n_neighbours = n_neighbours or self.n_neighbours
        result = np.zeros((len(dataframe), n_neighbours), dtype=np.int)
        for i in range(len(dataframe)):
            result[i] = self.index_builder.get_nns_by_vector(n=n_neighbours, search_k=search_k, vector=dataframe[self.features][i])
        return result

    def state_get(self):
        filename = tempfile.mktemp()
        self.index_builder.save(filename)
        with open(filename, 'rb') as f:
            data = f.read()
        return dict(tree_state=base64.encodebytes(data).decode('ascii'),
                    substate=super(ANNOYModel, self).state_get(),
                    n_dimensions=len(self.features))

    def state_set(self, state, trusted=True):
        super(ANNOYModel, self).state_set(state['substate'])
        data = base64.decodebytes(state['tree_state'].encode('ascii'))
        n_dimensions = state['n_dimensions']
        filename = tempfile.mktemp()
        with open(filename, 'wb') as f:
            f.write(data)
        self.index_builder = annoy.AnnoyIndex(n_dimensions)
        self.index_builder.load(filename)
        return self.index_builder


if __name__ == "__main__":
    ds = vaex.ml.datasets.load_iris()
    ds_train, ds_test = ds.ml.train_test_split()
    features = ds_train.column_names[:4]
    m = ANNOYModel(features=features, n_trees=50)
    m.fit(ds_train)
    m.predict(ds_test)
    m.transform(ds_test)



import vaex.dataframe
from vaex.serialize import register
import numpy as np
from . import generate
from .state import HasState
import traitlets
from vaex.utils import _ensure_strings_from_expressions

help_features = 'List of features to transform.'
help_prefix = 'Prefix for the names of the transformed features.'


def dot_product(a, b):
    products = ['%s * %s' % (ai, bi) for ai, bi in zip(a, b)]
    return ' + '.join(products)


@register
class StateTransfer(HasState):
    state = traitlets.Dict()

    def transform(self, df):
        copy = df.copy()
        self.state = dict(self.state, active_range=[copy._index_start, copy._index_end])
        copy.state_set(self.state)
        return copy


class Transformer(HasState):
    ''' Parent class for all of the transformers.
    '''
    features = traitlets.List(traitlets.Unicode(), help=help_features).tag(ui='SelectMultiple')

    def fit_transform(self, df):
        '''Fit and apply the transformer to the supplied DataFrame.

        :param df: A vaex DataFrame.

        :returns copy: A shallow copy of the DataFrame that includes the transformations.
        '''
        self.fit(df=df)
        return self.transform(df=df)


@register
@generate.register
class PCA(Transformer):
    '''Transform a set of features using a Principal Component Analysis.

    Example:

    >>> import vaex
    >>> df = vaex.from_arrays(x=[2,5,7,2,15], y=[-2,3,0,0,10])
    >>> df
     #   x   y
     0   2   -2
     1   5   3
     2   7   0
     3   2   0
     4   15  10
    >>> pca = vaex.ml.PCA(n_components=2, features=['x', 'y'])
    >>> pca.fit_transform(df)
     #    x    y       PCA_0      PCA_1
     0    2   -2    5.92532    0.413011
     1    5    3    0.380494  -1.39112
     2    7    0    0.840049   2.18502
     3    2    0    4.61287   -1.09612
     4   15   10  -11.7587    -0.110794

    '''
    # title = traitlets.Unicode(default_value='PCA', read_only=True).tag(ui='HTML')
    n_components = traitlets.Int(help='Number of components to retain. If None, all the components will be retained.').tag(ui='IntText')
    prefix = traitlets.Unicode(default_value="PCA_", help=help_prefix)
    progress = traitlets.CBool(default_value=False, help='If True, display a progressbar of the PCA fitting process.').tag(ui='Checkbox')
    eigen_vectors_ = traitlets.List(traitlets.List(traitlets.CFloat()), help='The eigen vectors corresponding to each feature').tag(output=True)
    eigen_values_ = traitlets.List(traitlets.CFloat(), help='The eigen values that correspond to each feature.').tag(output=True)
    means_ = traitlets.List(traitlets.CFloat(), help='The mean of each feature').tag(output=True)

    @traitlets.default('n_components')
    def get_n_components_default(self):
        return len(self.features)

    def fit(self, df):
        '''Fit the PCA model to the DataFrame.

        :param df: A vaex DataFrame.
        '''
        self.n_components = self.n_components or len(self.features)
        assert self.n_components >= 2, 'At least two features are required.'
        assert self.n_components <= len(self.features), 'Can not have more components than features.'
        C = df.cov(self.features, progress=self.progress)
        eigen_values, eigen_vectors = np.linalg.eigh(C)
        indices = np.argsort(eigen_values)[::-1]
        self.means_ = df.mean(self.features, progress=self.progress).tolist()
        self.eigen_vectors_ = eigen_vectors[:, indices].tolist()
        self.eigen_values_ = eigen_values[indices].tolist()

    def transform(self, df, n_components=None):
        '''Apply the PCA transformation to the DataFrame.

        :param df: A vaex DataFrame.
        :param n_components: The number of PCA components to retain.

        :return copy: A shallow copy of the DataFrame that includes the PCA components.
        :rtype: DataFrame
        '''
        n_components = n_components or self.n_components
        copy = df.copy()
        name_prefix_offset = 0
        eigen_vectors = np.array(self.eigen_vectors_)
        while self.prefix + str(name_prefix_offset) in copy.get_column_names(virtual=True, strings=True):
            name_prefix_offset += 1

        expressions = [copy[feature]-mean for feature, mean in zip(self.features, self.means_)]
        for i in range(n_components):
            v = eigen_vectors[:, i]
            expr = dot_product(expressions, v)
            name = self.prefix + str(i + name_prefix_offset)
            copy[name] = expr
        return copy


@register
@generate.register
class LabelEncoder(Transformer):
    '''Encode categorical columns with integer values between 0 and num_classes-1.

    Example:

    >>> import vaex
    >>> df = vaex.from_arrays(color=['red', 'green', 'green', 'blue', 'red'])
    >>> df
     #  color
     0  red
     1  green
     2  green
     3  blue
     4  red
    >>> encoder = vaex.ml.LabelEncoder(features=['color'])
    >>> encoder.fit_transform(df)
     #  color      label_encoded_color
     0  red                          2
     1  green                        1
     2  green                        1
     3  blue                         0
     4  red                          2
    '''
    # title = traitlets.Unicode(default_value='Label Encoder', read_only=True).tag(ui='HTML')
    prefix = traitlets.Unicode(default_value='label_encoded_', help=help_prefix).tag(ui='Text')
    labels_ = traitlets.Dict(default_value={}, allow_none=True, help='The encoded labels of each feature.').tag(output=True)
    allow_unseen = traitlets.Bool(default_value=False, allow_none=False, help='If True, unseen values will be \
                                  encoded with -1, otherwise an error is raised').tag(ui='Checkbox')

    def fit(self, df):
        '''Fit LabelEncoder to the DataFrame.

        :param df: A vaex DataFrame.
        '''

        for feature in self.features:
            labels = df[feature].unique().tolist()
            self.labels_[feature] = dict(zip(labels, np.arange(len(labels))))

    def transform(self, df):
        '''
        Transform a DataFrame with a fitted LabelEncoder.

        :param df: A vaex DataFrame.

        Returns:
        :return copy: A shallow copy of the DataFrame that includes the encodings.
        :rtype: DataFrame
        '''

        default_value = None
        if self.allow_unseen:
            default_value = -1

        copy = df.copy()
        for feature in self.features:
            name = self.prefix + feature
            copy[name] = copy[feature].map(mapper=self.labels_[feature], default_value=default_value)

        return copy


@register
@generate.register
class OneHotEncoder(Transformer):
    '''Encode categorical columns according ot the One-Hot scheme.

    Example:

    >>> import vaex
    >>> df = vaex.from_arrays(color=['red', 'green', 'green', 'blue', 'red'])
    >>> df
     #  color
     0  red
     1  green
     2  green
     3  blue
     4  red
    >>> encoder = vaex.ml.OneHotEncoder(features=['color'])
    >>> encoder.fit_transform(df)
     #  color      color_blue    color_green    color_red
     0  red                 0              0            1
     1  green               0              1            0
     2  green               0              1            0
     3  blue                1              0            0
     4  red                 0              0            1
    '''

    # title = Unicode(default_value='One-Hot Encoder', read_only=True).tag(ui='HTML')
    prefix = traitlets.Unicode(default_value='', help=help_prefix).tag(ui='Text')
    one = traitlets.Any(1, help='Value to encode when a category is present.')
    zero = traitlets.Any(0, help='Value to encode when category is absent.')
    uniques_ = traitlets.List(traitlets.List(), help='The unique elements found in each feature.').tag(output=True)

    def fit(self, df):
        '''Fit OneHotEncoder to the DataFrame.

        :param df: A vaex DataFrame.
        '''

        uniques = []
        for i in self.features:
            expression = _ensure_strings_from_expressions(i)
            unique = df.unique(expression)
            unique = np.sort(unique)  # this can/should be optimized with @delay
            uniques.append(unique.tolist())
        self.uniques_ = uniques

    def transform(self, df):
        '''Transform a DataFrame with a fitted OneHotEncoder.

        :param df: A vaex DataFrame.
        :return: A shallow copy of the DataFrame that includes the encodings.
        :rtype: DataFrame
        '''
        copy = df.copy()
        # for each feature, add a virtual column for each unique entry
        for i, feature in enumerate(self.features):
            for j, value in enumerate(self.uniques_[i]):
                column_name = self.prefix + feature + '_' + str(value)
                copy.add_virtual_column(column_name, 'where({feature} == {value}, {one}, {zero})'.format(
                                        feature=feature, value=repr(value), one=self.one, zero=self.zero))
        return copy


@register
@generate.register
class FrequencyEncoder(Transformer):
    '''Encode categorical columns by the frequency of their respective samples.

    Example:

    >>> import vaex
    >>> df = vaex.from_arrays(color=['red', 'green', 'green', 'blue', 'red', 'green'])
    >>> df
     #  color
     0  red
     1  green
     2  green
     3  blue
     4  red
    >>> encoder = vaex.ml.FrequencyEncoder(features=['color'])
    >>> encoder.fit_transform(df)
     #  color      frequency_encoded_color
     0  red                       0.333333
     1  green                     0.5
     2  green                     0.5
     3  blue                      0.166667
     4  red                       0.333333
     5  green                     0.5
    '''
    prefix = traitlets.Unicode(default_value='frequency_encoded_', help=help_prefix).tag(ui='Text')
    unseen = traitlets.Enum(values=['zero', 'nan'], default_value='nan', help='Strategy to deal with unseen values.')
    mappings_ = traitlets.Dict()

    def fit(self, df):
        '''Fit FrequencyEncoder to the DataFrame.

        :param df: A vaex DataFrame.
        '''
        # number of samples
        nsamples = len(df)

        # Encoding
        for feature in self.features:
            self.mappings_[feature] = dict(df[feature].value_counts() / nsamples)

    def transform(self, df):
        '''Transform a DataFrame with a fitted FrequencyEncoder.

        :param df: A vaex DataFrame.
        :return: A shallow copy of the DataFrame that includes the encodings.
        :rtype: DataFrame
        '''
        copy = df.copy()
        default_value = {'zero': 0., 'nan': np.nan}[self.unseen]
        for feature in self.features:
            name = self.prefix + feature
            expression = copy[feature].map(self.mappings_[feature], nan_value=np.nan, missing_value=np.nan, default_value=default_value, allow_missing=True)

            copy[name] = expression
        return copy


@register
@generate.register
class StandardScaler(Transformer):
    '''Standardize features by removing thir mean and scaling them to unit variance.

    Example:

    >>> import vaex
    >>> df = vaex.from_arrays(x=[2,5,7,2,15], y=[-2,3,0,0,10])
    >>> df
     #    x    y
     0    2   -2
     1    5    3
     2    7    0
     3    2    0
     4   15   10
    >>> scaler = vaex.ml.StandardScaler(features=['x', 'y'])
    >>> scaler.fit_transform(df)
     #    x    y    standard_scaled_x    standard_scaled_y
     0    2   -2            -0.876523            -0.996616
     1    5    3            -0.250435             0.189832
     2    7    0             0.166957            -0.522037
     3    2    0            -0.876523            -0.522037
     4   15   10             1.83652              1.85086
    '''
    # title = Unicode(default_value='Standard Scaler', read_only=True).tag(ui='HTML')
    prefix = traitlets.Unicode(default_value="standard_scaled_", help=help_prefix).tag(ui='Text')
    with_mean = traitlets.CBool(default_value=True, help='If True, remove the mean from each feature.').tag(ui='Checkbox')
    with_std = traitlets.CBool(default_value=True, help='If True, scale each feature to unit variance.').tag(ui='Checkbox')
    mean_ = traitlets.List(traitlets.CFloat(), help='The mean of each feature').tag(output=True)
    std_ = traitlets.List(traitlets.CFloat(), help='The standard deviation of each feature.').tag(output=True)

    def fit(self, df):
        '''
        Fit StandardScaler to the DataFrame.

        :param df: A vaex DataFrame.
        '''
        mean = df.mean(self.features, delay=True)
        std = df.std(self.features, delay=True)

        @vaex.delayed
        def assign(mean, std):
            self.mean_ = mean.tolist()
            self.std_ = std.tolist()

        assign(mean, std)
        df.execute()

    def transform(self, df):
        '''
        Transform a DataFrame with a fitted StandardScaler.

        :param df: A vaex DataFrame.

        :returns copy: a shallow copy of the DataFrame that includes the scaled features.
        :rtype: DataFrame
        '''

        copy = df.copy()
        for i in range(len(self.features)):
            name = self.prefix+self.features[i]
            expression = copy[self.features[i]]
            if self.with_mean:
                expression = expression - self.mean_[i]
            if self.with_std:
                expression = expression / self.std_[i]
            copy[name] = expression
        return copy


@register
@generate.register
class MinMaxScaler(Transformer):
    '''Will scale a set of features to a given range.

    Example:

    >>> import vaex
    >>> df = vaex.from_arrays(x=[2,5,7,2,15], y=[-2,3,0,0,10])
    >>> df
     #    x    y
     0    2   -2
     1    5    3
     2    7    0
     3    2    0
     4   15   10
    >>> scaler = vaex.ml.MinMaxScaler(features=['x', 'y'])
    >>> scaler.fit_transform(df)
     #    x    y    minmax_scaled_x    minmax_scaled_y
     0    2   -2           0                  0
     1    5    3           0.230769           0.416667
     2    7    0           0.384615           0.166667
     3    2    0           0                  0.166667
     4   15   10           1                  1
    '''
    # title = Unicode(default_value='MinMax Scaler', read_only=True).tag(ui='HTML')
    feature_range = traitlets.Tuple(default_value=(0, 1), help='The range the features are scaled to.').tag().tag(ui='FloatRangeSlider')
    prefix = traitlets.Unicode(default_value="minmax_scaled_", help=help_prefix).tag(ui='Text')
    fmax_ = traitlets.List(traitlets.CFloat(), help='The minimum value of a feature.').tag(output=True)
    fmin_ = traitlets.List(traitlets.CFloat(), help='The maximum value of a feature.').tag(output=True)

    def fit(self, df):
        '''
        Fit MinMaxScaler to the DataFrame.

        :param df: A vaex DataFrame.
        '''

        assert len(self.feature_range) == 2, 'feature_range must have 2 elements only'
        minmax = df.minmax(self.features)
        self.fmin_ = minmax[:, 0].tolist()
        self.fmax_ = minmax[:, 1].tolist()

    def transform(self, df):
        '''
        Transform a DataFrame with a fitted MinMaxScaler.

        :param df: A vaex DataFrame.

        :return copy: a shallow copy of the DataFrame that includes the scaled features.
        :rtype: DataFrame
        '''

        copy = df.copy()

        for i in range(len(self.features)):
            name = self.prefix + self.features[i]
            a = self.feature_range[0]
            b = self.feature_range[1]
            expr = copy[self.features[i]]
            expr = (b-a)*(expr-self.fmin_[i])/(self.fmax_[i]-self.fmin_[i]) + a
            copy[name] = expr
        return copy


@register
@generate.register
class MaxAbsScaler(Transformer):
    ''' Scale features by their maximum absolute value.

    Example:

    >>> import vaex
    >>> df = vaex.from_arrays(x=[2,5,7,2,15], y=[-2,3,0,0,10])
    >>> df
     #    x    y
     0    2   -2
     1    5    3
     2    7    0
     3    2    0
     4   15   10
    >>> scaler = vaex.ml.MaxAbsScaler(features=['x', 'y'])
    >>> scaler.fit_transform(df)
     #    x    y    absmax_scaled_x    absmax_scaled_y
     0    2   -2           0.133333               -0.2
     1    5    3           0.333333                0.3
     2    7    0           0.466667                0
     3    2    0           0.133333                0
     4   15   10           1                       1
    '''
    prefix = traitlets.Unicode(default_value="absmax_scaled_", help=help_prefix).tag(ui='Text')
    absmax_ = traitlets.List(traitlets.CFloat(), help='Tha maximum absolute value of a feature.').tag(output=True)

    def fit(self, df):
        '''
        Fit MinMaxScaler to the DataFrame.

        :param df: A vaex DataFrame.
        '''

        absmax = df.max(['abs(%s)' % k for k in self.features]).tolist()
        # Check if the absmax_ value is 0, in which case replace with 1
        self.absmax_ = [value if value != 0 else 1 for value in absmax]

    def transform(self, df):
        '''
        Transform a DataFrame with a fitted MaxAbsScaler.

        :param df: A vaex DataFrame.

        :return copy: a shallow copy of the DataFrame that includes the scaled features.
        :rtype: DataFrame
        '''

        copy = df.copy()
        for i in range(len(self.features)):
            name = self.prefix + self.features[i]
            expr = copy[self.features[i]]
            expr = expr / self.absmax_[i]
            copy[name] = expr
        return copy


@register
@generate.register
class RobustScaler(Transformer):
    ''' The RobustScaler removes the median and scales the data according to a
    given percentile range. By default, the scaling is done between the 25th and
    the 75th percentile. Centering and scaling happens independently for each
    feature (column).

    Example:

    >>> import vaex
    >>> df = vaex.from_arrays(x=[2,5,7,2,15], y=[-2,3,0,0,10])
    >>> df
     #    x    y
     0    2   -2
     1    5    3
     2    7    0
     3    2    0
     4   15   10
    >>> scaler = vaex.ml.MaxAbsScaler(features=['x', 'y'])
    >>> scaler.fit_transform(df)
     #    x    y    robust_scaled_x    robust_scaled_y
     0    2   -2       -0.333686             -0.266302
     1    5    3       -0.000596934           0.399453
     2    7    0        0.221462              0
     3    2    0       -0.333686              0
     4   15   10        1.1097                1.33151
    '''
    with_centering = traitlets.CBool(default_value=True, help='If True, remove the median.').tag(ui='Checkbox')
    with_scaling = traitlets.CBool(default_value=True, help='If True, scale each feature between the specified percentile range.').tag(ui='Checkbox')
    percentile_range = traitlets.Tuple(default_value=(25, 75), help='The percentile range to which to scale each feature to.').tag().tag(ui='FloatRangeSlider')
    prefix = traitlets.Unicode(default_value="robust_scaled_", help=help_prefix).tag(ui='Text')
    center_ = traitlets.List(traitlets.CFloat(), default_value=None, help='The median of each feature.').tag(output=True)
    scale_ = traitlets.List(traitlets.CFloat(), default_value=None, help='The percentile range for each feature.').tag(output=True)

    def fit(self, df):
        '''
        Fit RobustScaler to the DataFrame.

        :param df: A vaex DataFrame.
        '''

        # check the quantile range
        q_min, q_max = self.percentile_range
        if not 0 <= q_min <= q_max <= 100:
            raise ValueError('Invalid percentile range: %s' % (str(self.percentile_range)))

        if self.with_centering:
            self.center_ = df.percentile_approx(expression=self.features, percentage=50).tolist()

        if self.with_scaling:
            self.scale_ = (df.percentile_approx(expression=self.features, percentage=q_max) - df.percentile_approx(expression=self.features, percentage=q_min)).tolist()

    def transform(self, df):
        '''
        Transform a DataFrame with a fitted RobustScaler.

        :param df: A vaex DataFrame.

        :returns copy: a shallow copy of the DataFrame that includes the scaled features.
        :rtype: DataFrame
        '''

        copy = df.copy()
        for i in range(len(self.features)):
            name = self.prefix+self.features[i]
            expr = copy[self.features[i]]
            if self.with_centering:
                expr = expr - self.center_[i]
            if self.with_scaling:
                expr = expr / self.scale_[i]
            copy[name] = expr
        return copy



import vaex
import pygbm
import numpy as np
import vaex.serialize
from .. import state
import traitlets
import pygbm.binning
import contextlib
import cloudpickle as pickle
import base64

import pygbm.gradient_boosting
from .. import generate

@contextlib.contextmanager
def using_vaex(features):
    """Temporary monkey patches pygbm"""
    class VaexBinMapper(traitlets.HasTraits):
        max_bins = traitlets.CInt(255)
        random_state = traitlets.Any()
        subsample = traitlets.CInt(int(1e5))

        def fit(self, dataframe_wrapper):
            dataframe = dataframe_wrapper.ds
            self.bin_thresholds_ = []
            for feature in features:
                X = dataframe[feature].values.reshape((-1, 1)).astype(np.float32)
                midpoints = pygbm.binning.find_binning_thresholds(
                    X, self.max_bins, subsample=self.subsample,
                    random_state=self.random_state)[0]
                self.bin_thresholds_.append(midpoints)
            self.bin_thresholds_

        def transform(self, dataframe_wrapper):
            dataframe = dataframe_wrapper.ds
            N = len(dataframe)
            M = len(features)
            # fortran order so 1 column is contiguous in memory
            binned = np.zeros((N, M), dtype=np.uint8, order='F')
            for m, feature in enumerate(features):
                X = dataframe[feature].values.reshape((-1, 1)).astype(np.float32)
                binned1 = pygbm.binning.map_to_bins(X, binning_thresholds=self.bin_thresholds_)
                assert binned1.shape[1] == 1
                binned[:,m] = binned1[:,0]
            return binned

        def fit_transform(self, X):
            self.fit(X)
            return self.transform(X)
    try:
        check_X_y = pygbm.gradient_boosting.check_X_y
        BinMapper = pygbm.gradient_boosting.BinMapper

        pygbm.gradient_boosting.BinMapper = VaexBinMapper
        pygbm.gradient_boosting.check_X_y = lambda *x, **kwargs: x
        yield
    finally:
        pygbm.gradient_boosting.BinMapper = BinMapper
        pygbm.gradient_boosting.check_X_y = check_X_y



class DataFrameWrapper:
    def __init__(self, ds):
        self.ds = ds

    @property
    def nbytes(self):
        return self.ds.byte_size()


@vaex.serialize.register
@generate.register
class PyGBMModel(state.HasState):

    features = traitlets.List(traitlets.Unicode())
    num_round = traitlets.CInt()
    param = traitlets.Dict()
    prediction_name = traitlets.Unicode(default_value='pygbm_prediction')
    learning_rate = traitlets.Float(0.1)
    max_iter = traitlets.Int(10)
    max_bins = traitlets.Int(255)
    max_leaf_nodes = traitlets.Int(31)
    random_state = traitlets.Int(0)
    verbose = traitlets.Int(1)
    prediction_name = traitlets.Unicode(default_value='pygbm_prediction')


    def fit(self, dataframe, label):
        self.pygbm_model = pygbm.GradientBoostingMachine(
                            learning_rate=self.learning_rate,
                            max_iter=self.max_iter,
                            max_bins=self.max_bins,
                            max_leaf_nodes=self.max_leaf_nodes,
                            random_state=self.random_state,
                            scoring=None,
                            verbose=self.verbose,
                            validation_split=None)
        if not hasattr(label, 'values'):
            label = dataframe[label]
        y = label.values.astype(np.float32)
        with using_vaex(self.features):
            dsw = DataFrameWrapper(dataframe)
            self.pygbm_model.fit(dsw, y)

    def predict(self, dataframe):
        data = np.vstack([dataframe[k].values for k in self.features]).T
        return self.pygbm_model.predict(data)

    def __call__(self, *args):
        data = np.vstack([arg.astype(np.float32) for arg in args]).T.copy()
        return self.pygbm_model.predict(data)

    def transform(self, dataframe):
        copy = dataframe.copy()
        lazy_function = copy.add_function('pygbm_prediction_function', self)
        expression = lazy_function(*self.features)
        copy.add_virtual_column(self.prediction_name, expression, unique=False)
        return copy

    def state_get(self):
        return dict(tree_state=base64.encodebytes(pickle.dumps(self.pygbm_model)).decode('ascii'),
                    substate=super(PyGBMModel, self).state_get())

    def state_set(self, state, trusted=True):
        super(PyGBMModel, self).state_set(state['substate'])
        if trusted is False:
            raise ValueError("Will not unpickle data when source is not trusted")
        self.pygbm_model = pickle.loads(base64.decodebytes(state['tree_state'].encode('ascii')))

@vaex.serialize.register
class PyGBMClassifier(PyGBMModel):
    def __call__(self, *args):
        return np.argmax(super(PyGBMClassifier, self).__call__(*args), axis=1)
    def predict(self, dataframe, copy=False):
        return np.argmax(super(PyGBMClassifier, self).predict(dataframe, copy=copy), axis=1)



import os
import vaex
import numpy as np


def load_iris():
    '''Load and return the iris dataset (classification).

    The iris dataset is a classic and very easy multi-class classification dataset.

    =================   ==============
    Classes                          3
    Samples per class               50
    Samples total                  150
    Dimensionality                   4
    Features            real, positive
    =================   ==============

    Example:
    ========

    >>> import vaex.ml
    >>> df = vaex.ml.datasets.load_iris()
    >>> df.describe()
    '''
    dirname = os.path.dirname(__file__)
    return vaex.open(os.path.join(dirname, 'iris.hdf5'))


def _iris(name, iris_previous, N):
    filename = os.path.join(vaex.utils.get_private_dir('data'), name + '.hdf5')
    if os.path.exists(filename):
        return vaex.open(filename)
    else:
        iris = iris_previous()
        repeat = int(np.ceil(N / len(iris)))
        ds = vaex.dataset.DatasetConcatenated([iris] * repeat)
        ds.export_hdf5(filename)
        return vaex.open(filename)


def iris_subsample(N, error_percentage=5, ds=None):
    '''Returns the iris set repeated so it include ~1e4 rows'''
    # return _iris_subsample('iris_1e4', iris, int(1e4))
    ds = ds or load_iris()
    ds_out = None
    repeats = int(np.ceil(N / len(ds)))
    for feature in ds.get_column_names():
        if feature in ['random_index']:
            continue
        data = ds[feature].values
        min, max = ds.minmax(feature)
        error = (max - min)/100*error_percentage
        data_out = np.repeat(data, repeats)
        if feature not in ['class_']:
            data_out += np.random.random(len(data_out)) * error
        if ds_out is None:
            ds_out = vaex.from_arrays(feature=data_out)
        else:
            ds_out.add_column(feature, data_out)
    return ds_out


def load_iris_1e4():
    '''Returns the iris set repeated so it include ~1e4 rows'''
    return _iris('iris_1e4', load_iris, int(1e4))


def load_iris_1e5():
    '''Returns the iris set repeated so it include ~1e5 rows'''
    return _iris('iris_1e5', load_iris_1e4, int(1e5))


def load_iris_1e6():
    '''Returns the iris set repeated so it include ~1e6 rows'''

    return _iris('iris_1e6', load_iris_1e5, int(1e6))


def load_iris_1e7():
    '''Returns the iris set repeated so it include ~1e7 rows'''
    return _iris('iris_1e7', load_iris_1e6, int(1e7))


def load_iris_1e8():
    '''Returns the iris set repeated so it include ~1e8 rows'''
    return _iris('iris_1e8', load_iris_1e7, int(1e8))


def load_iris_1e9():
    '''Returns the iris set repeated so it include ~1e8 rows'''
    return _iris('iris_1e9', load_iris_1e8, int(1e9))


def load_titanic():
    '''
    Returns the classic Titanic dataset.

    Description of the columns can be found in dataset.description.

    Example:
    ========

    >>> import vaex.ml
    >>> df = vaex.mk.dataset.load_titanic()
    >>> print(df.description)
    >>> df.describe()
    '''
    dirname = os.path.dirname(__file__)
    return vaex.open(os.path.join(dirname, 'titanic.hdf5'))



import vaex.utils
import traitlets
import pickle
import base64


def object_pickle(trait_name, value, state_obj):
    binary_data = pickle.dumps(value)
    data = base64.encodebytes(binary_data).decode('ascii')
    return data


def object_unpickle(trait_name, data, state_obj):
    binary_data = base64.decodebytes(data.encode('ascii'))
    value = pickle.loads(binary_data)
    return value


serialize_pickle = {
    'to_json': object_pickle,
    'from_json': object_unpickle
}


def default_to_json(trait_name, value, state_obj):
    return value


def default_from_json(trait_name, data, state_obj):
    return data


class HasState(traitlets.HasTraits):

    @classmethod
    def state_from(cls, state, trusted=True):
        obj = cls()
        obj.state_set(state)
        return obj

    def state_get(self):
        state = {}
        for name in self.trait_names():
            to_json = self.trait_metadata(name, 'to_json', default_to_json)
            value = to_json(name, getattr(self, name), self)
            state[name] = value
        return state

    def state_set(self, state, trusted=True):
        for name in self.trait_names():
            if name in state:
                from_json = self.trait_metadata(name, 'from_json', default_from_json)
                value = from_json(name, state[name], self)
                setattr(self, name, value)

    def state_write(self, f):
        vaex.utils.write_json_or_yaml(f, self.state_get())

    def state_load(self, f):
        state = vaex.utils.read_json_or_yaml(f)
        self.state_set(state)



import base64
import tempfile
import traitlets

import vaex
import vaex.serialize
from . import state
from . import generate

import numpy as np
import catboost


@vaex.serialize.register
@generate.register
class CatBoostModel(state.HasState):
    '''The CatBoost algorithm.

    This class provides an interface to the CatBoost aloritham.
    CatBoost is a fast, scalable, high performance Gradient Boosting on
    Decision Trees library, used for ranking, classification, regression and
    other machine learning tasks. For more information please visit
    https://github.com/catboost/catboost

    Example:

    >>> import vaex
    >>> import vaex.ml.catboost
    >>> df = vaex.ml.datasets.load_iris()
    >>> features = ['sepal_width', 'petal_length', 'sepal_length', 'petal_width']
    >>> df_train, df_test = vaex.ml.train_test_split(df)
    >>> params = {
        'leaf_estimation_method': 'Gradient',
        'learning_rate': 0.1,
        'max_depth': 3,
        'bootstrap_type': 'Bernoulli',
        'objective': 'MultiClass',
        'eval_metric': 'MultiClass',
        'subsample': 0.8,
        'random_state': 42,
        'verbose': 0}
    >>> booster = vaex.ml.catboost.CatBoostModel(features=features, num_boost_round=100, params=params)
    >>> booster.fit(df_train, 'class_')
    >>> df_train = booster.transform(df_train)
    >>> df_train.head(3)
    #    sepal_length    sepal_width    petal_length    petal_width    class_  catboost_prediction
    0             5.4            3               4.5            1.5         1  [0.00615039 0.98024259 0.01360702]
    1             4.8            3.4             1.6            0.2         0  [0.99034267 0.00526382 0.0043935 ]
    2             6.9            3.1             4.9            1.5         1  [0.00688241 0.95190908 0.04120851]
    >>> df_test = booster.transform(df_test)
    >>> df_test.head(3)
    #    sepal_length    sepal_width    petal_length    petal_width    class_  catboost_prediction
    0             5.9            3               4.2            1.5         1  [0.00464228 0.98883351 0.00652421]
    1             6.1            3               4.6            1.4         1  [0.00350424 0.9882139  0.00828186]
    2             6.6            2.9             4.6            1.3         1  [0.00325705 0.98891631 0.00782664]
    '''

    features = traitlets.List(traitlets.Unicode(), help='List of features to use when fitting the CatBoostModel.')
    num_boost_round = traitlets.CInt(default_value=None, allow_none=True, help='Number of boosting iterations.')
    params = traitlets.Dict(help='A dictionary of parameters to be passed on to the CatBoostModel model.')
    pool_params = traitlets.Dict(default_value={}, help='A dictionary of parameters to be passed to the Pool data object construction')
    prediction_name = traitlets.Unicode(default_value='catboost_prediction', help='The name of the virtual column housing the predictions.')
    prediction_type = traitlets.Enum(values=['Probability', 'Class', 'RawFormulaVal'], default_value='Probability',
                                     help='The form of the predictions. Can be "RawFormulaVal", "Probability" or "Class".')

    def __call__(self, *args):
        data2d = np.vstack([arg.astype(np.float64) for arg in args]).T.copy()
        dmatrix = catboost.Pool(data2d, **self.pool_params)
        return self.booster.predict(dmatrix, prediction_type=self.prediction_type)

    def transform(self, df):
        '''Transform a DataFrame such that it contains the predictions of the CatBoostModel in form of a virtual column.

        :param df: A vaex DataFrame. It should have the same columns as the DataFrame used to train the model.

        :return copy: A shallow copy of the DataFrame that includes the CatBoostModel prediction as a virtual column.
        :rtype: DataFrame
        '''
        copy = df.copy()
        lazy_function = copy.add_function('catboost_prediction_function', self)
        expression = lazy_function(*self.features)
        copy.add_virtual_column(self.prediction_name, expression, unique=False)
        return copy

    def fit(self, df, target, evals=None, early_stopping_rounds=None,
            verbose_eval=None, plot=False, **kwargs):
        '''Fit the CatBoostModel model given a DataFrame.
        This method accepts all key word arguments for the catboost.train method.

        :param df: A vaex DataFrame containing the training features.
        :param target: The column name of the target variable.
        :param evals: A list of DataFrames to be evaluated during training.
        This allows user to watch performance on the validation sets.
        :param int early_stopping_rounds: Activates early stopping.
        :param bool verbose_eval: Requires at least one item in *evals*.
        If *verbose_eval* is True then the evaluation metric on the validation set is printed at each boosting stage.
        :param bool plot: if True, display an interactive widget in the Jupyter
        notebook of how the train and validation sets score on each boosting iteration.
        '''
        # Ensure strings
        target = vaex.dataframe._ensure_string_from_expression(target)

        data = df[self.features].values
        target_data = df[target].values
        dtrain = catboost.Pool(data=data, label=target_data, **self.pool_params)
        if evals is not None:
            for i, item in enumerate(evals):
                data = item[self.features].values
                target_data = item[target].values
                evals[i] = catboost.Pool(data=data, label=target_data, **self.pool_params)

        # This does the actual training/fitting of the catboost model
        self.booster = catboost.train(params=self.params,
                                      dtrain=dtrain,
                                      num_boost_round=self.num_boost_round,
                                      evals=evals,
                                      early_stopping_rounds=early_stopping_rounds,
                                      verbose_eval=verbose_eval,
                                      plot=plot,
                                      **kwargs)

    def predict(self, df, **kwargs):
        '''Provided a vaex DataFrame, get an in-memory numpy array with the predictions from the CatBoostModel model.
        This method accepts the key word arguments of the predict method from catboost.

        :param df: a vaex DataFrame

        :returns: A in-memory numpy array containing the CatBoostModel predictions.
        :rtype: numpy.array
        '''
        data = df[self.features].values
        dmatrix = catboost.Pool(data, **self.pool_params)
        return self.booster.predict(dmatrix, prediction_type=self.prediction_type, **kwargs)

    def state_get(self):
        filename = tempfile.mktemp()
        self.booster.save_model(filename)
        with open(filename, 'rb') as f:
            data = f.read()
        return dict(tree_state=base64.encodebytes(data).decode('ascii'),
                    substate=super(CatBoostModel, self).state_get())

    def state_set(self, state, trusted=True):
        super(CatBoostModel, self).state_set(state['substate'])
        data = base64.decodebytes(state['tree_state'].encode('ascii'))
        filename = tempfile.mktemp()
        with open(filename, 'wb') as f:
            f.write(data)
        self.booster = catboost.CatBoost().load_model(fname=filename)



import warnings

import vaex
import vaex.dataframe
from . import datasets
from .pipeline import Pipeline

from vaex.utils import InnerNamespace
from vaex.utils import _ensure_strings_from_expressions


class DataFrameAccessorML(object):
    def __init__(self, df):
        self.df = df

    def pca(self, n_components=2, features=None, prefix='PCA_', progress=False):
        '''Requires vaex.ml: Create :class:`vaex.ml.transformations.PCA` and fit it.

        :param n_components: Number of components to retain. If None, all the components will be retained.
        :param features: List of features to transform.
        :param prefix: Prefix for the names of the transformed features.
        :param progress: If True, display a progressbar of the PCA fitting process.
        '''
        features = features or self.df.get_column_names()
        features = _ensure_strings_from_expressions(features)
        pca = PCA(n_components=n_components, features=features, progress=progress)
        pca.fit(self.df)
        return pca


    def label_encoder(self, features=None, prefix='label_encoded_', allow_unseen=False):
        '''Requires vaex.ml: Create :class:`vaex.ml.transformations.LabelEncoder` and fit it.

        :param features: List of features to encode.
        :param prefix: Prefix for the names of the encoded features.
        :param allow_unseen: If True, encode unseen value as -1, otherwise an error is raised.
        '''
        features = features or self.df.get_column_names()
        features = _ensure_strings_from_expressions(features)
        label_encoder = LabelEncoder(features=features, prefix=prefix, allow_unseen=allow_unseen)
        label_encoder.fit(self.df)
        return label_encoder


    def one_hot_encoder(self, features=None, one=1, zero=0, prefix=''):
        '''Requires vaex.ml: Create :class:`vaex.ml.transformations.OneHotEncoder` and fit it.

        :param features: List of features to encode.
        :param one: What value to use instead of "1".
        :param zero: What value to use instead of "0".
        :param prefix: Prefix for the names of the encoded features.
        :returns one_hot_encoder: vaex.ml.transformations.OneHotEncoder object.
        '''
        if features is None:
            raise ValueError('Please give at least one categorical feature.')
        features = _ensure_strings_from_expressions(features)
        one_hot_encoder = OneHotEncoder(features=features, one=one, zero=zero, prefix=prefix)
        one_hot_encoder.fit(self.df)
        return one_hot_encoder


    def frequency_encoder(self, features=None, unseen='nan', prefix='frequency_encoded_'):
        '''
        Requires vaex.ml: Create :class:`vaex.ml.transformations.FrequencyEncoder` and fit it.

        :param features: List of features to encode.
        :param unseen: Strategy to deal with unseen values. Accepted arguments are "zero" or "nan".
        :param prefix: Prefix for the names of the encoded features.
        '''
        features = features or self.df.get_column_names()
        features = _ensure_strings_from_expressions(features)
        freq_encoder = FrequencyEncoder(features=features, prefix=prefix)
        freq_encoder.fit(self.df)
        return freq_encoder


    def standard_scaler(self, features=None, with_mean=True, with_std=True, prefix='standard_scaled_'):
        '''Requires vaex.ml: Create :class:`vaex.ml.transformations.StandardScaler` and fit it.

        :param features: List of features to scale.
        :param with_mean: If True, remove the mean from each feature.
        :param with_std: If True, scale each feature to unit variance.
        :param prefix: Prefix for the names of the scaled features.
        '''
        features = features or self.df.get_column_names()
        features = _ensure_strings_from_expressions(features)
        standard_scaler = StandardScaler(features=features, with_mean=with_mean, with_std=with_std, prefix=prefix)
        standard_scaler.fit(self.df)
        return standard_scaler


    def minmax_scaler(self, features=None, feature_range=[0, 1], prefix='minmax_scaled_'):
        '''Requires vaex.ml: Create :class:`vaex.ml.transformations.MinMaxScaler` and fit it.

        :param features: List of features to scale.
        :param feature_range: The range the features are scaled to.
        :param prefix: Prefix for the names of the scaled features.
        '''
        features = features or self.df.get_column_names()
        features = _ensure_strings_from_expressions(features)
        minmax_scaler = MinMaxScaler(features=features, feature_range=feature_range, prefix=prefix)
        minmax_scaler.fit(self.df)
        return minmax_scaler


    def xgboost_model(self, target, num_boost_round, features=None, params={}, prediction_name='xgboost_prediction'):
        '''Requires vaex.ml: create a XGBoost model and train/fit it.

        :param target: Target to train/fit on.
        :param num_boost_round: Number of rounds.
        :param features: List of features to train on.
        :return vaex.ml.xgboost.XGBModel: Fitted XGBoost model.
        '''
        from .xgboost import XGBoostModel
        dataframe = self.df
        features = features or self.df.get_column_names()
        features = _ensure_strings_from_expressions(features)
        booster = XGBoostModel(prediction_name=prediction_name,
                            num_boost_round=num_boost_round,
                            features=features,
                            params=params)
        booster.fit(dataframe, target)
        return booster


    def lightgbm_model(self, target, num_boost_round, features=None, copy=False, params={},
                    prediction_name='lightgbm_prediction'):
        '''Requires vaex.ml: create a lightgbm model and train/fit it.

        :param target: The target variable to predict.
        :param num_boost_round: Number of boosting iterations.
        :param features: List of features to train on.
        :param bool copy: Copy data or use the modified xgboost library for efficient transfer.
        :return vaex.ml.lightgbm.LightGBMModel: Fitted LightGBM model.
        '''
        from .lightgbm import LightGBMModel
        dataframe = self.df
        features = features or self.df.get_column_names(virtual=True)
        features = _ensure_strings_from_expressions(features)

        booster = LightGBMModel(prediction_name=prediction_name,
                                num_boost_round=num_boost_round,
                                features=features,
                                params=params)
        booster.fit(dataframe, target, copy=copy)
        return booster


    def catboost_model(self, target, num_boost_round, features=None, params=None, prediction_name='catboost_prediction'):
        '''Requires vaex.ml: create a CatBoostModel model and train/fit it.

        :param target: Target to train/fit on
        :param num_boost_round: Number of rounds
        :param features: List of features to train on
        :return vaex.ml.catboost.CatBoostModel: Fitted CatBoostModel model
        '''
        from .catboost import CatBoostModel
        dataframe = self.df
        features = features or self.df.get_column_names()
        features = _ensure_strings_from_expressions(features)
        booster = CatBoostModel(prediction_name=prediction_name,
                                num_boost_round=num_boost_round,
                                features=features,
                                params=params)
        booster.fit(dataframe, target)
        return booster


    def pygbm_model(self, label, max_iter, features=None, param={}, classifier=False, prediction_name='pygbm_prediction', **kwargs):
        '''Requires vaex.ml: create a pygbm model and train/fit it.

        :param label: Label to train/fit on
        :param max_iter: Max number of iterations/trees
        :param features: List of features to train on
        :param bool classifier: If true, return a the classifier (will use argmax on the probabilities)
        :return vaex.ml.pygbm.PyGBMModel or vaex.ml.pygbm.PyGBMClassifier: Fitted PyGBM model
        '''
        from .incubator.pygbm import PyGBMModel, PyGBMClassifier
        dataframe = self.df
        features = features or self.df.get_column_names()
        features = _ensure_strings_from_expressions(features)
        cls = PyGBMClassifier if classifier else PyGBMModel
        b = cls(prediction_name=prediction_name, max_iter=max_iter, features=features, param=param, **kwargs)
        b.fit(dataframe, label)
        return b


    def state_transfer(self):
        from .transformations import StateTransfer
        state = self.df.state_get()
        state.pop('active_range')  # we are not interested in this..
        return StateTransfer(state=state)


    def train_test_split(self, test_size=0.2, strings=True, virtual=True, verbose=True):
        '''Will split the DataFrame in train and test part, assuming it is shuffled.

        :param test_size: The fractional size of the test set.
        :param strings: If True, the output DataFrames will also contain string columns, if any.
        :param virtual: If True, the output DataFrames will also contain virtual contain, if any.
        :param verbose: If True, print warnings to screen.
        '''
        if verbose:
            warnings.warn('Make sure the DataFrame is shuffled')
        initial = None
        try:
            assert self.df.filtered is False, 'Filtered DataFrames are not yet supported.'
            # full_length = len(self)
            df = self.df.trim()
            initial = self.df.get_active_range()
            df.set_active_fraction(test_size)
            test = df.trim()
            __, end = df.get_active_range()
            df.set_active_range(end, df.length_original())
            train = df.trim()
        finally:
            if initial is not None:
                df.set_active_range(*initial)
        return train, test


from .transformations import PCA
from .transformations import StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler
from .transformations import LabelEncoder, OneHotEncoder, FrequencyEncoder



import pickle
import base64
import tempfile

import traitlets
import numpy as np

import vaex
import vaex.serialize
from vaex.ml import state
from vaex.ml import generate
from vaex.ml.state import serialize_pickle


@vaex.serialize.register
@generate.register
class SKLearnPredictor(state.HasState):
    '''This class wraps any scikit-learn estimator (a.k.a predictor) making it a vaex pipeline object.

    By wrapping any scikit-learn estimators with this class, it becoes a vaex
    pipeline object. Thus, it can take full advantage of the serialization and
    pipeline system of vaex. One can use the `predict` method to get a numpy
    array as an output of a fitted estimator, or the `transform` method do add
    such a prediction to a vaex DataFrame as a virtual column.

    Note that a full memory copy of the data used is created when the `fit` and
    `predict` are called. The `transform` method is evaluated lazily.

    The scikit-learn estimators themselves are not modified at all, they are
    taken from your local installation of scikit-learn.

    Example:

    >>> import vaex.ml
    >>> from vaex.ml.sklearn import SKLearnPredictor
    >>> from sklearn.linear_model import LinearRegression
    >>> df = vaex.ml.datasets.load_iris()
    >>> features = ['sepal_width', 'petal_length', 'sepal_length']
    >>> df_train, df_test = vaex.ml.train_test_split(df)
    >>> model = SKLearnPredictor(model=LinearRegression(), features=features, prediction_name='pred')
    >>> model.fit(df_train, df_train.petal_width)
    >>> df_train = model.transform(df_train)
    >>> df_train.head(3)
     #    sepal_length    sepal_width    petal_length    petal_width    class_      pred
     0             5.4            3               4.5            1.5         1  1.64701
     1             4.8            3.4             1.6            0.2         0  0.352236
     2             6.9            3.1             4.9            1.5         1  1.59336
    >>> df_test = model.transform(df_test)
    >>> df_test.head(3)
     #    sepal_length    sepal_width    petal_length    petal_width    class_     pred
     0             5.9            3               4.2            1.5         1  1.39437
     1             6.1            3               4.6            1.4         1  1.56469
     2             6.6            2.9             4.6            1.3         1  1.44276
    '''

    model = traitlets.Any(default_value=None, allow_none=True, help='A scikit-learn estimator.').tag(**serialize_pickle)
    features = traitlets.List(traitlets.Unicode(), help='List of features to use.')
    prediction_name = traitlets.Unicode(default_value='prediction', help='The name of the virtual column housing the predictions.')

    def __call__(self, *args):
        X = np.vstack([arg.astype(np.float64) for arg in args]).T.copy()
        return self.model.predict(X)

    def predict(self, df):
        '''Get an in-memory numpy array with the predictions of the SKLearnPredictor.self

        :param df: A vaex DataFrame, containing the input features.
        :returns: A in-memory numpy array containing the SKLearnPredictor predictions.
        :rtype: numpy.array
        '''
        data = df[self.features].values
        return self.model.predict(data)

    def transform(self, df):
        '''Transform a DataFrame such that it contains the predictions of the SKLearnPredictor.
        in form of a virtual column.

        :param df: A vaex DataFrame.

        :return copy: A shallow copy of the DataFrame that includes the SKLearnPredictor prediction as a virtual column.
        :rtype: DataFrame
        '''
        copy = df.copy()
        lazy_function = copy.add_function('sklearn_prediction_function', self)
        expression = lazy_function(*self.features)
        copy.add_virtual_column(self.prediction_name, expression, unique=False)
        return copy

    def fit(self, df, target, **kwargs):
        '''Fit the SKLearnPredictor to the DataFrame.

        :param df: A vaex DataFrame containing the features on which to train the model.
        :param target: The name of the column containing the target variable.
        '''
        X = df[self.features].values
        y = df.evaluate(target)
        self.model.fit(X=X, y=y, **kwargs)



from collections import defaultdict
import os
import re
import sys

from jinja2 import Template
import traitlets

import vaex

# registry = {}
registry = []


def register(cls):
    docstring_args_template = Template("""
{% for arg in docargs %}
:param {{ arg.name }}: {{ arg.help }}{% endfor %}
""")

    template_method = Template("""

# from {{ module }} import *
# import traitlets

def {{ method_name }}(self, {{ signature }}):
    obj = {{module}}.{{ class_name }}({{ args }})
    obj.fit(self{{ extra_fit }})
    return obj

import vaex.dataframe
# vaex.dataframe.DataFrame.ml._add({{ method_name }}={{ method_name }})

def __init__(self, {{ full_signature }}):
    \"\"\"
    {{ docstring_args }}
    \"\"\"
    given_kwargs = {key:value for key, value in dict({{ full_args }}).items() if value is not traitlets.Undefined}

    super({{module}}.{{ class_name }}, self).__init__(**given_kwargs)

cls.__init__ = __init__
cls.__signature__ = __init__
del __init__
    """)
    traits = {key: value for key, value in cls.class_traits().items()
                if 'output' not in value.metadata}
    traits_nodefault = {key: value for key, value in traits.items()
                        if value.default_value is traitlets.Undefined}
    traits_default = {key: value for key, value in traits.items()
                        if key not in traits_nodefault}
    signature_list = ["{name}".format(name=name, value=value.default_value)
                        for name, value in traits_nodefault.items()]
    signature_list.extend(["{name}={value!r}".format(name=name, value=value.default_value)
                            for name, value in traits_default.items()])
    signature = ", ".join(signature_list)
    args = ", ".join(["{name}={name}".format(name=name, value=value)
                        for name, value in traits.items()])

    signature_list = ["{name}={value!r}".format(name=name, value=value.default_value)
                        for name, value in traits.items()]
    full_signature = ", ".join(signature_list)
    full_args = args

    method_name = camel_to_underscore(cls.__name__)
    module = cls.__module__
    class_name = cls.__name__

    doctraits = {name: trait for name, trait in traits.items() if 'help' in trait.metadata}
    docargs = [{'name': name, 'help': trait.metadata['help']} for name, trait in doctraits.items()]

    kwargs = dict(locals())
    docstring_args = docstring_args_template.render(**kwargs)

    kwargs = dict(locals())
    code = template_method.render(**kwargs)
    exec(code, globals(), locals())
    return cls

def camel_to_underscore(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()



import traitlets
from . import generate
import ipywidgets as widgets
from vaex.serialize import register


@register
class Widgetize(widgets.VBox):
    def __init__(self, pipeline_object, dataframe):
        '''
        Docstring
        :params:
        :param pipeline_object: the vaex pipeline element object
        :param dataframe: a vaex DataFrame on which the pipeline object will operate
        '''
        super(Widgetize, self).__init__()

        self.pipeline_object = pipeline_object
        self.dataframe = dataframe

        ui_dict_ = {
            'Checkbox': widgets.Checkbox,
            'Text': widgets.Text,
            'IntText': widgets.IntText,
            'FloatText': widgets.FloatText,
            'Button': widgets.Button,
            'SelectMultiple': widgets.SelectMultiple,
            'FloatRangeSlider': widgets.FloatRangeSlider,
            'HTML': widgets.HTML,
        }

        widget_list = []
        for trait_name in pipeline_object.trait_names():
            metadata = pipeline_object.traits()[trait_name].metadata
            # only traits with ui metadata will show up in the ui
            if 'ui' in metadata:
                widget_class = ui_dict_[metadata['ui']]
                widget = widget_class(description=trait_name)
                widget_list.append(widget)

                if isinstance(widget, widgets.SelectMultiple):
                    widget.options = self.dataframe.get_column_names(virtual=True, strings=True)
                    traitlets.link((self.pipeline_object, trait_name), (widget, 'value'))
                else:
                    traitlets.link((self.pipeline_object, trait_name), (widget, 'value'))

        fit_button = widgets.Button(description='Fit', button_style='info')
        fit_button.on_click(lambda button: self.pipeline_object.fit(self.dataframe))
        # add it to the 'form'.
        widget_list.append(fit_button)

        # Adding all the  widget elements to the parent object
        self.children = widget_list



from vaex.serialize import to_dict, from_dict
import vaex.utils


class Pipeline(list):

    def save(self, f):
        states = [to_dict(k) for k in self]
        vaex.utils.write_json_or_yaml(f, states)

    def load(self, f):
        states = vaex.utils.read_json_or_yaml(f)
        objects = [from_dict(k) for k in states]
        del self[:]  # TODOPY2: replace by .clear
        self.extend(objects)

    def predict(self, dataframe):
        assert len(self) > 0, "cannot predict when pipeline is empty"
        # apply transformations for all but the last
        transforms = self[:-1]
        for t in transforms:
            dataframe = t.transform(dataframe)
        # last does predictions
        return self[-1].predict(dataframe)

    def transform(self, dataframe):
        assert len(self) > 0, "cannot predict when pipeline is empty"
        # apply transformations for all but the last
        for t in self:
            dataframe = t.transform(dataframe)
        return dataframe



import os
import imp
from setuptools import setup

dirname = os.path.dirname(__file__)
path_version = os.path.join(dirname, 'vaex/ml/_version.py')
version = imp.load_source('version', path_version)

name        = 'vaex'
author      = 'Jovan Veljanoski'
author_email= 'jovan.veljanoski@gmail.com'
license     = 'MIT'
version     = version.__version__
url         = 'https://www.github.com/vaexio/vaex'
install_requires_ml = ['vaex-core>=1.0.0,<2', 'numba', 'traitlets','jinja2']

setup(name=name + '-ml',
      version=version,
      description='Machine learning support for vaex',
      url=url,
      author=author,
      author_email=author_email,
      install_requires=install_requires_ml,
      license=license,
      packages=['vaex.ml', 'vaex.ml.incubator', 'vaex.ml.datasets'],
      include_package_data=True,
      zip_safe=False,
      entry_points={'vaex.dataframe.accessor': ['ml = vaex.ml:DataFrameAccessorML']}
)



import numpy as np
import traitlets
import sklearn.linear_model as lin

import vaex
import vaex.serialize
from . import state
from . import generate


def listify(l):
    if hasattr(l, 'tolist'):
        return l.tolist()
    else:
        try:
            return [listify(k) for k in l]
        except TypeError:
            return l


class _LinearBase(state.HasState):
    features = traitlets.List(traitlets.Unicode())
    binned = traitlets.CBool(True)
    shape = traitlets.CInt(64)
    limits = traitlets.List(traitlets.List(traitlets.CFloat(), allow_none=True),
                            allow_none=True).tag(output=True)
    fit_intercept = traitlets.CBool(True)
    coef_ = traitlets.Union(
                            [traitlets.List(traitlets.CFloat()),
                             traitlets.List(traitlets.List(traitlets.CFloat()))]
                            ).tag(output=True)
    # intercept_ = traitlets.List(traitlets.CFloat).tag(output=True)
    intercept_ = traitlets.Union([traitlets.CFloat(),
                                 traitlets.List(traitlets.CFloat())]).tag(output=True)
    _sk_params = traitlets.Any()
    prediction_name = traitlets.Unicode(default_value='linear_prediction')

    def transform(self, dataset):
        ds = dataset.copy()
        expression = self.coef_[0] * ds[self.features[0]]
        for coef, feature in zip(self.coef_, self.features[1:]):
            expression = expression + coef * ds[feature]
        expression = self.intercept_ + expression
        ds.add_virtual_column(self.prediction_name, expression, unique=False)
        return ds

    def fit(self, dataset, y_expression, progress=False):
        assert len(set(self.features)) == len(self.features), "duplicate features"
        if not self.binned:
            X = np.array(dataset[self.features])
            y = dataset.evaluate(y_expression)
            m = self._make_model()
            m.fit(X, y)
            self.coef_ = m.coef_.tolist()
            self.intercept_ = m.intercept_.tolist()
        else:
            limits = self.limits
            if limits == []:
                limits = None
            binby = self.features + [y_expression]
            limits, shapes = dataset.limits(binby, limits, shape=self.shape)
            self.limits = listify(limits)
            counts = dataset.count(binby=binby, limits=limits, shape=shapes)
            mask = counts > 0

            def coordinates(expression, limits, shape):
                if dataset.is_category(expression):
                    return np.arange(dataset.category_count(expression))
                else:
                    return dataset.bin_centers(expression, limits, shape)
            centers = [coordinates(expression, l, shape) for expression, l, shape
                       in zip(binby, self.limits, shapes)]
            # l = ds.bin_centers('y', limits[1], shape)
            centers = np.meshgrid(*centers, indexing='ij')
            centers = [c[mask] for c in centers]
            # m = lin.LinearRegression(fit_intercept=self.fit_intercept)
            m = self._make_model()
            X = np.array(centers[:-1]).reshape(-1, len(self.features))
            y = centers[-1].reshape(-1)
            weights = counts[mask]
            m.fit(X, y, sample_weight=weights)
            self.coef_ = m.coef_.tolist()
            self.intercept_ = m.intercept_.tolist()
        # self._sk_params = m.get_params()
        self.last_model = m

    def predict(self, dataset):
        X = np.array(dataset[self.features])
        return self.last_model.predict(X)


@generate.register
@vaex.serialize.register
class LogisticRegression(_LinearBase):
    prediction_name = traitlets.Unicode(default_value='logit_prediction')

    def _make_model(self):
        return lin.LogisticRegression(fit_intercept=self.fit_intercept)


@generate.register
@vaex.serialize.register
class LinearRegression(_LinearBase):
    prediction_name = traitlets.Unicode(default_value='linear_prediction')

    def _make_model(self):
        return lin.LinearRegression(fit_intercept=self.fit_intercept)





import base64
import tempfile
import traitlets

import xgboost
import numpy as np

import vaex
from . import state
from . import generate
import vaex.serialize


@vaex.serialize.register
@generate.register
class XGBoostModel(state.HasState):
    '''The XGBoost algorithm.

    XGBoost is an optimized distributed gradient boosting library designed to be
    highly efficient, flexible and portable. It implements machine learning
    algorithms under the Gradient Boosting framework. XGBoost provides a parallel
    tree boosting (also known as GBDT, GBM) that solve many data science
    problems in a fast and accurate way.
    (https://github.com/dmlc/xgboost)

    Example:

    >>> import vaex
    >>> import vaex.ml.xgboost
    >>> df = vaex.ml.datasets.load_iris()
    >>> features = ['sepal_width', 'petal_length', 'sepal_length', 'petal_width']
    >>> df_train, df_test = vaex.ml.train_test_split(df)
    >>> params = {
        'max_depth': 5,
        'learning_rate': 0.1,
        'objective': 'multi:softmax',
        'num_class': 3,
        'subsample': 0.80,
        'colsample_bytree': 0.80,
        'silent': 1}
    >>> booster = vaex.ml.xgboost.XGBoostModel(features=features, num_boost_round=100, params=params)
    >>> booster.fit(df_train, 'class_')
    >>> df_train = booster.transform(df_train)
    >>> df_train.head(3)
    #    sepal_length    sepal_width    petal_length    petal_width    class_    xgboost_prediction
    0             5.4            3               4.5            1.5         1                     1
    1             4.8            3.4             1.6            0.2         0                     0
    2             6.9            3.1             4.9            1.5         1                     1
    >>> df_test = booster.transform(df_test)
    >>> df_test.head(3)
    #    sepal_length    sepal_width    petal_length    petal_width    class_    xgboost_prediction
    0             5.9            3               4.2            1.5         1                     1
    1             6.1            3               4.6            1.4         1                     1
    2             6.6            2.9             4.6            1.3         1                     1
    '''

    features = traitlets.List(traitlets.Unicode(), help='List of features to use when fitting the XGBoostModel.')
    num_boost_round = traitlets.CInt(help='Number of boosting iterations.')
    params = traitlets.Dict(help='A dictionary of parameters to be passed on to the XGBoost model.')
    prediction_name = traitlets.Unicode(default_value='xgboost_prediction', help='The name of the virtual column housing the predictions.')

    def __call__(self, *args):
        data2d = np.vstack([arg.astype(np.float64) for arg in args]).T.copy()
        dmatrix = xgboost.DMatrix(data2d)
        return self.booster.predict(dmatrix)

    def transform(self, df):
        '''Transform a DataFrame such that it contains the predictions of the XGBoostModel in form of a virtual column.

        :param df: A vaex DataFrame. It should have the same columns as the DataFrame used to train the model.
        :return copy: A shallow copy of the DataFrame that includes the XGBoostModel prediction as a virtual column.
        :rtype: DataFrame
        '''
        copy = df.copy()
        lazy_function = copy.add_function('xgboost_prediction_function', self)
        expression = lazy_function(*self.features)
        copy.add_virtual_column(self.prediction_name, expression, unique=False)
        return copy

    def fit(self, df, target, evals=(), early_stopping_rounds=None,
            evals_result=None, verbose_eval=False, **kwargs):
        '''Fit the XGBoost model given a DataFrame.

        This method accepts all key word arguments for the xgboost.train method.

        :param df: A vaex DataFrame containing the training features.
        :param target: The column name of the target variable.
        :param evals: A list of pairs (DataFrame, string).
            List of items to be evaluated during training, this allows user to watch performance on the validation set.
        :param int early_stopping_rounds: Activates early stopping.
            Validation error needs to decrease at least every *early_stopping_rounds* round(s) to continue training.
            Requires at least one item in *evals*. If there's more than one, will use the last. Returns the model
            from the last iteration (not the best one).
        :param dict evals_result: A dictionary storing the evaluation results of all the items in *evals*.
        :param bool verbose_eval: Requires at least one item in *evals*.
            If *verbose_eval* is True then the evaluation metric on the validation set is printed at each boosting stage.
        '''
        data = df[self.features].values
        target_data = df.evaluate(target)
        dtrain = xgboost.DMatrix(data, target_data)
        if evals is not None:
            evals = [list(elem) for elem in evals]
            for item in evals:
                data = item[0][self.features].values
                target_data = item[0].evaluate(target)
                item[0] = xgboost.DMatrix(data, target_data)
        else:
            evals = ()

        # This does the actual training / fitting of the xgboost model
        self.booster = xgboost.train(params=self.params,
                                     dtrain=dtrain,
                                     num_boost_round=self.num_boost_round,
                                     evals=evals,
                                     early_stopping_rounds=early_stopping_rounds,
                                     evals_result=evals_result,
                                     verbose_eval=verbose_eval,
                                     **kwargs)

    def predict(self, df, **kwargs):
        '''Provided a vaex DataFrame, get an in-memory numpy array with the predictions from the XGBoost model.
        This method accepts the key word arguments of the predict method from XGBoost.

        :returns: A in-memory numpy array containing the XGBoostModel predictions.
        :rtype: numpy.array
        '''
        data = df[self.features].values
        dmatrix = xgboost.DMatrix(data)
        return self.booster.predict(dmatrix, **kwargs)

    def state_get(self):
        filename = tempfile.mktemp()
        self.booster.save_model(filename)
        with open(filename, 'rb') as f:
            data = f.read()
        return dict(tree_state=base64.encodebytes(data).decode('ascii'),
                    substate=super(XGBoostModel, self).state_get())

    def state_set(self, state, trusted=True):
        super(XGBoostModel, self).state_set(state['substate'])
        data = base64.decodebytes(state['tree_state'].encode('ascii'))
        filename = tempfile.mktemp()
        with open(filename, 'wb') as f:
            f.write(data)
        self.booster = xgboost.Booster(model_file=filename)



import ctypes
import math
import warnings

import vaex
import lightgbm
import numpy as np
import tempfile
import base64
import vaex.serialize
from . import state
import traitlets
from . import generate


lib = lightgbm.basic._LIB


@vaex.serialize.register
@generate.register
class LightGBMModel(state.HasState):
    '''The LightGBM algorithm.

    This class provides an interface to the LightGBM algorithm, with some optimizations
    for better memory efficiency when training large datasets. The algorithm itself is
    not modified at all.

    LightGBM is a fast gradient boosting algorithm based on decision trees and is
    mainly used for classification, regression and ranking tasks. It is under the
    umbrella of the Distributed Machine Learning Toolkit (DMTK) project of Microsoft.
    For more information, please visit https://github.com/Microsoft/LightGBM/.

    Example:

    >>> import vaex.ml
    >>> import vaex.ml.lightgbm
    >>> df = vaex.ml.datasets.load_iris()
    >>> features = ['sepal_width', 'petal_length', 'sepal_length', 'petal_width']
    >>> df_train, df_test = vaex.ml.train_test_split(df)
    >>> params = {
        'boosting': 'gbdt',
        'max_depth': 5,
        'learning_rate': 0.1,
        'application': 'multiclass',
        'num_class': 3,
        'subsample': 0.80,
        'colsample_bytree': 0.80}
    >>> booster = vaex.ml.lightgbm.LightGBMModel(features=features, num_boost_round=100, params=params)
    >>> booster.fit(df_train, 'class_')
    >>> df_train = booster.transform(df_train)
    >>> df_train.head(3)
     #    sepal_width    petal_length    sepal_length    petal_width    class_    lightgbm_prediction
     0            3               4.5             5.4            1.5         1    [0.00165619 0.98097899 0.01736482]
     1            3.4             1.6             4.8            0.2         0    [9.99803930e-01 1.17346471e-04 7.87235133e-05]
     2            3.1             4.9             6.9            1.5         1    [0.00107541 0.9848717  0.01405289]
    >>> df_test = booster.transform(df_test)
    >>> df_test.head(3)
     #    sepal_width    petal_length    sepal_length    petal_width    class_    lightgbm_prediction
     0            3               4.2             5.9            1.5         1    [0.00208904 0.9821348  0.01577616]
     1            3               4.6             6.1            1.4         1    [0.00182039 0.98491357 0.01326604]
     2            2.9             4.6             6.6            1.3         1    [2.50915444e-04 9.98431777e-01 1.31730785e-03]
    '''
    features = traitlets.List(traitlets.Unicode(), help='List of features to use when fitting the LightGBMModel.')
    num_boost_round = traitlets.CInt(help='Number of boosting iterations.')
    params = traitlets.Dict(help='parameters to be passed on the to the LightGBM model.')
    prediction_name = traitlets.Unicode(default_value='lightgbm_prediction', help='The name of the virtual column housing the predictions.')

    def __call__(self, *args):
        data2d = np.vstack([arg.astype(np.float64) for arg in args]).T.copy()
        return self.booster.predict(data2d)

    def transform(self, df):
        '''Transform a DataFrame such that it contains the predictions of the LightGBMModel
        in form of a virtual column.

        :param df: A vaex DataFrame.

        :return copy: A shallow copy of the DataFrame that includes the LightGBMModel prediction as a virtual column.
        :rtype: DataFrame
        '''
        copy = df.copy()
        lazy_function = copy.add_function('lightgbm_prediction_function', self)
        expression = lazy_function(*self.features)
        copy.add_virtual_column(self.prediction_name, expression, unique=False)
        return copy

    def fit(self, df, target, valid_sets=None, valid_names=None,
            early_stopping_rounds=None, evals_result=None, verbose_eval=None,
            copy=False, **kwargs):
        """Fit the LightGBMModel to the DataFrame.

        The model will train until the validation score stops improving.
        Validation score needs to improve at least every *early_stopping_rounds* rounds
        to continue training. Requires at least one validation DataFrame, metric
        specified. If there's more than one, will check all of them, but the
        training data is ignored anyway. If early stopping occurs, the model
        will add ``best_iteration`` field to the booster object.
        :param dict evals_result: A dictionary storing the evaluation results of all *valid_sets*.
        :param bool verbose_eval: Requires at least one item in *evals*.
        If *verbose_eval* is True then the evaluation metric on the validation
        set is printed at each boosting stage.
        :param bool copy: (default, False) If True, make an in memory copy of the data before passing it to LightGBMModel.


        :param df: A vaex DataFrame.
        :param target: The name of the column containing the target variable.
        :param list valid_sets: A list of DataFrames to be used for validation.
        :param list valid_names: A list of strings to label the validation sets.
        :param early_stopping_rounds int: Activates early stopping.

        """

        if copy:
            dtrain = lightgbm.Dataset(df[self.features].values, df.evaluate(target))
            if valid_sets is not None:
                for i, item in enumerate(valid_sets):
                    valid_sets[i] = lightgbm.Dataset(item[self.features].values, item[target].values)
            else:
                valid_sets = ()
        else:
            dtrain = VaexDataset(df, target, features=self.features)
            if valid_sets is not None:
                for i, item in enumerate(valid_sets):
                    # valid_sets[i] = VaexDataset(item, target, features=self.features)
                    valid_sets[i] = lightgbm.Dataset(item[self.features].values, item[target].values)
                    warnings.warn('''Validation sets do not obey the `copy=False` argument.
                                  A standard in-memory copy is made for each validation set.''', UserWarning)
            else:
                valid_sets = ()

        self.booster = lightgbm.train(params=self.params,
                                      train_set=dtrain,
                                      num_boost_round=self.num_boost_round,
                                      valid_sets=valid_sets,
                                      valid_names=valid_names,
                                      early_stopping_rounds=early_stopping_rounds,
                                      evals_result=evals_result,
                                      verbose_eval=verbose_eval,
                                      **kwargs)

    def predict(self, df, **kwargs):
        '''Get an in-memory numpy array with the predictions of the LightGBMModel on a vaex DataFrame.
        This method accepts the key word arguments of the predict method from LightGBM.

        :param df: A vaex DataFrame.

        :returns: A in-memory numpy array containing the LightGBMModel predictions.
        :rtype: numpy.array
        '''
        # TODO: we want to go multithreaded/parallel/chunks
        return self.booster.predict(df[self.features].values)

    def state_get(self):
        filename = tempfile.mktemp()
        self.booster.save_model(filename)
        with open(filename, 'rb') as f:
            data = f.read()
        return dict(tree_state=base64.encodebytes(data).decode('ascii'),
                    substate=super(LightGBMModel, self).state_get())

    def state_set(self, state, trusted=True):
        super(LightGBMModel, self).state_set(state['substate'])
        data = base64.decodebytes(state['tree_state'].encode('ascii'))
        filename = tempfile.mktemp()
        with open(filename, 'wb') as f:
            f.write(data)
        self.booster = lightgbm.Booster(model_file=filename)


class VaexDataset(lightgbm.Dataset):
    def __init__(self, df, label=None, features=None, blocksize=100*1000, sample_count=10*100, params={}):
        super(VaexDataset, self).__init__(None)
        self.df = df
        self.features = features or self.df.get_column_names(virtual=True)
        assert len(set(self.features)) == len(self.features), "using duplicate features"
        self.blocksize = blocksize

        self.data_references = {}

        self.c_features = (ctypes.c_char_p * len(self.features))()
        self.c_features[:] = [k.encode() for k in self.features]

        self.handle = ctypes.c_void_p()

        row_count = len(self.df)
        ncol = len(self.features)
        num_sample_row = min(sample_count, row_count)

        data_list = [self.df.evaluate(k, i1=0, i2=num_sample_row).astype(np.float64) for k in self.features]
        data_pointers = [k.ctypes.data_as(ctypes.POINTER(ctypes.c_double)) for k in data_list]
        sample_data = (ctypes.POINTER(ctypes.c_double) * ncol)(*data_pointers)

        indices = np.arange(num_sample_row, dtype=np.int32)
        indices_pointers = [indices.ctypes.data_as(ctypes.POINTER(ctypes.c_int)) for k in range(ncol)]
        sample_indices = (ctypes.POINTER(ctypes.c_int) * ncol)(*indices_pointers)
        num_per_col = (ctypes.c_int * ncol)(*((num_sample_row,) * ncol))
        parameters = ctypes.c_char_p(lightgbm.basic.param_dict_to_str(params).encode())

        lightgbm.basic._safe_call(lib.LGBM_DatasetCreateFromSampledColumn(sample_data,
                                                                          sample_indices,
                                                                          ctypes.c_uint(ncol),
                                                                          num_per_col,
                                                                          ctypes.c_uint(num_sample_row),
                                                                          ctypes.c_uint(row_count),
                                                                          parameters,
                                                                          ctypes.byref(self.handle)))

        blocks = int(math.ceil(row_count / blocksize))
        dtype = np.float64
        for i in range(blocks):
            i1 = i * blocksize
            i2 = min(row_count, (i+1) * blocksize)
            data = np.array([df.evaluate(k, i1=i1, i2=i2).astype(dtype) for k in self.features]).T.copy()
            ctypemap = {np.float64: ctypes.c_double, np.float32: ctypes.c_float}
            capi_typemap = {np.float64: lightgbm.basic.C_API_DTYPE_FLOAT64, np.float32: lightgbm.basic.C_API_DTYPE_FLOAT32}
            lightgbm.basic._safe_call(lib.LGBM_DatasetPushRows(self.handle,
                                                               data.ctypes.data_as(ctypes.POINTER(ctypemap[dtype])),
                                                               ctypes.c_uint(capi_typemap[dtype]),
                                                               ctypes.c_uint32(i2-i1),
                                                               ctypes.c_uint32(ncol),
                                                               ctypes.c_uint32(i1),
                                                               ))

        if label is not None:
            self.label_data = self.df.evaluate(label)
            self.set_label(self.label_data)


if __name__ == "__main__":
    df = vaex.ml.iris()
    features = ['sepal_length', 'sepal_width', 'petal_length', 'petal_width']
    # ldf = VaexDataset(df, 'class_', features=features)
    ldf = lightgbm.Dataset(np.array(df[features]), df.data.class_)
    param = {'num_leaves': 31, 'num_trees': 100, 'objective': 'softmax', 'num_class': 3}
    num_boost_round = 30
    booster = lightgbm.train(param, ldf, num_boost_round)
    print(booster.predict(np.array(df[features])))
